#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>

#define APPLICATION_SNTP_PORT 123

/*Идентификаторы классов сообщений*/
enum {
	APPLICATION_PROCESS_START = 0,
	APPLICATION_PROCESS_END,
	APPLICATION_PROCESS_ERROR
};

/*Представление пакета протокола SNTP*/
typedef struct __attribute__((packed)) {
	guint8  mode                    : 3; /*Индикатор коррекции (3 бита)*/
	guint8  version                 : 3; /*Номер версии (3 бита)*/
	guint8  leapinfo                : 2; /*Режим (2 бита)*/
	guint8  stratum;                     /*Страта*/
	gint8   poll;                        /*Интервал опроса*/
	gint8   precision;                   /*Точность*/
	guint32 root_delay;                  /*Задержка обращения*/
	guint32 root_dispersion;             /*Относительная ошибка часов сервера*/
	gchar   reference_id[4];             /*Идентификатор источника*/
	guint32 reference_timestamp_hi;      /*Время обновления (секунды)*/
	guint32 reference_timestamp_lo;      /*Время обновления (доли секунды)*/
	guint32 originate_timestamp_hi;      /*Начальное время (секунды)*/
	guint32 originate_timestamp_lo;      /*Начальное время (доли секунды)*/
	guint32 receive_timestamp_hi;        /*Время приема (секунды)*/
	guint32 receive_timestamp_lo;        /*Время приема (доли секунды)*/
	guint32 transmit_timestamp_hi;       /*Время отправки (секунды)*/
	guint32 transmit_timestamp_lo;       /*Время отправки (доли секунды)*/
} sntpdg_t;

/*Данные приложения*/
typedef struct {
	/*Объект приложения*/
	GtkApplication *app;
	/*Указатель на структуру, представляющую пакет протокола SNTP*/
	sntpdg_t *sntpdata;
	/*Идентификатор таймера обновления значений времени*/
	guint timeoutsrc;
	/*Флаг синхронизации времени с сетевым сервером*/
	gboolean timesynced;
	/*Разница во времени между локальным компьютером и удаленным сервером*/
	GTimeSpan timeoffset;
	/*Виджет вывода времени локального компьютера*/
	GtkWidget *systime_value;
	/*Виджет вывода времени, синхронизированного с удаленным сервером*/
	GtkWidget *synctime_value;
	/*Виджет основного окна приложения*/
	GtkWidget *mainwin; 
	/*Виджет для ввода имени домена*/
	GtkWidget *server_combo;
	/*Виджет кнопки для получения информации*/
	GtkWidget *sync_button;
	/*Виджет поля для редактирования текста*/
	GtkWidget *whois_textview;
	/*Виджет строки состояния*/
	GtkWidget *statusbar;
	/*Идентификатор последнего сообщения виджета строки состояния*/
	guint statusmsgid;
	/*Идентификатор контекста виджета строки состояния*/
	guint statusconid;
	/*Виджет индикатора выполнения задачи*/
	GtkWidget *spinner;
} appdata;

/*Функция изменения состояния графического интерфейса и вывода соответствующего сообщения
  в строке состояния*/
static void application_process_ui_change(appdata *data, guint processid, GError *error)
{
	gchar *statusmsg;
	
	g_assert(data != NULL);
	
	switch (processid) {
		case APPLICATION_PROCESS_START:
			/*Начало процесса синхронизации времени*/
			statusmsg = g_strdup_printf("Синхронизация времени: %s", gtk_combo_box_text_get_active_text(GTK_COMBO_BOX_TEXT(data->server_combo)));
			gtk_widget_set_sensitive(data->sync_button, FALSE);
			gtk_widget_set_sensitive(data->server_combo, FALSE);
			gtk_widget_show(data->spinner);
			gtk_spinner_start(GTK_SPINNER(data->spinner));
			break;
		case APPLICATION_PROCESS_END:
			/*Окончание процесса псинхронизации времени*/
			statusmsg = g_strdup_printf("Время синхронизировано");
			gtk_widget_set_sensitive(data->sync_button, TRUE);
			gtk_widget_set_sensitive(data->server_combo, TRUE);
			gtk_spinner_stop(GTK_SPINNER(data->spinner));
			gtk_widget_hide(data->spinner);
			break;
		case APPLICATION_PROCESS_ERROR:
			/*Ошибка в процессе синхронизации времени*/
			if (error != NULL) {
				statusmsg = g_strdup_printf("Ошибка: %s", error->message);
				g_error_free(error);
			} else {
				statusmsg = g_strdup_printf("Неизвестная ошибка");
			}
			gtk_widget_set_sensitive(data->sync_button, TRUE);
			gtk_widget_set_sensitive(data->server_combo, TRUE);
			gtk_spinner_stop(GTK_SPINNER(data->spinner));
			gtk_widget_hide(data->spinner);
			break;
		default:
			/*Неизвестное действие*/
			statusmsg = g_strdup_printf("Неизвестное действие");
			break;
	}
	
	/*Удаление предыдущей строки*/
	if (data->statusmsgid != 0) {
		gtk_statusbar_remove(GTK_STATUSBAR(data->statusbar), data->statusconid, data->statusmsgid);
	}
	/*Вывод строки*/
	data->statusmsgid = gtk_statusbar_push(GTK_STATUSBAR(data->statusbar), data->statusconid, statusmsg);
	/*Освобождение памяти*/
	g_free(statusmsg);
}

/*Функция обратного вызова для получения данных от сервера*/
static gboolean application_read_callback(GSocket *socket, GIOCondition cond, gpointer user_data)
{
	appdata *data;
	GError *error;
	gssize len;
	GDateTime *datetime, *synctime;
	GTimeVal tv;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	if (cond == G_IO_IN) {
		/*Проверка выполнения установленного условия доступности данных (G_IO_IN)*/
		error = NULL;
		/*Получение дейтаграммы от сервера синхронизации времени*/
		len = g_socket_receive_from(socket, NULL, (gchar *)data->sntpdata, sizeof(sntpdg_t), NULL, &error);
		if (len > 0) {
			/*Преобразование переданной сервером метки времени в секунды и микросекунды*/
			tv.tv_sec  = GUINT32_FROM_BE(data->sntpdata->transmit_timestamp_hi) - 2208988800;
			tv.tv_usec = GUINT32_FROM_BE(data->sntpdata->transmit_timestamp_lo) / 4295;
			/*Преобразование секунд и микросекунд в синхронизированную метку времени*/
			synctime = g_date_time_new_from_timeval_local((const GTimeVal *)&tv);
			if (synctime != NULL) {
				/*Получение текущей метки времени*/
				datetime = g_date_time_new_now_local();
				if (datetime != NULL) {
					/*Вычисление разницы между текущей и синхронизированной метками времени*/
					data->timeoffset = g_date_time_difference(synctime, datetime);
					/*Установка флага синхронизации времени*/
					data->timesynced = TRUE;
					/*Завершение приема и обработки данных*/
					application_process_ui_change(data, APPLICATION_PROCESS_END, NULL);
					/*Освобождение памяти, используемой для хранения меток времени*/
					g_date_time_unref(datetime);
					g_date_time_unref(synctime);
				} else {
					/*Вывод сообщения об ошибке в случае невозможности вычисления текущей метки времени*/
					application_process_ui_change(data, APPLICATION_PROCESS_ERROR, NULL);
					g_date_time_unref(synctime);
				}
			} else {
				/*Вывод сообщения об ошибке в случае невозможности вычисления синхронизированной метки времени*/
				application_process_ui_change(data, APPLICATION_PROCESS_ERROR, NULL);
			}
		} else {
			/*Вывод сообщения об ошибке в случае невозможности приема данных от сервера*/
			application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
		}
	} else {
		/*Обработка стандартных условий, соотвествующих ошибкам (G_IO_ERR, G_IO_HUP) */
		application_process_ui_change(data, APPLICATION_PROCESS_ERROR, NULL);
	}
	
	/*Закрытие сетевого сокета и освобождение памяти, использованной для хранения структуры, представляющей
	  пакет протокола SNTP*/
	g_object_unref(socket);
	g_free(data->sntpdata);
	return FALSE;
}

/*Функция обратного вызова для обработки результатов разрешения доменных имен*/
static void application_resolver_lookup_callback(GObject *source_object, GAsyncResult *res, gpointer user_data)
{
	appdata *data;
	GError *error = NULL;
	GList *addresses;
	GInetAddress *inetaddress;
	GSocketAddress *socketaddress;
	GSocket *socket;
	GSource *source;
	GDateTime *datetime;
	GTimeVal tv;
	gssize len;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Получение списка IP-адресов, соотвествующих заданному доменному имени*/
	addresses = g_resolver_lookup_by_name_finish(G_RESOLVER(source_object), res, &error);
	
	if (addresses != NULL) {
		/*Извлечение первого IP-адреса из списка (индекс первого объекта списка равен 0)*/
		inetaddress = G_INET_ADDRESS(g_list_nth_data(addresses, 0));
		/*Преобразование объекта IP-адреса в объект адреса сетевого сокета с указанием номера порта*/
		socketaddress = g_inet_socket_address_new(inetaddress, APPLICATION_SNTP_PORT);
		/*Создание объекта сетевого сокета для работы с протоколом UDP*/
		socket = g_socket_new(g_inet_address_get_family(inetaddress), G_SOCKET_TYPE_DATAGRAM, G_SOCKET_PROTOCOL_DEFAULT, &error);
		if (socket != NULL) {
			/*Установка ограничения по времени для операций с сетевым сокетом (равного 5 секундам)*/
			g_socket_set_timeout(socket, 5);
			/*Резервирование памяти для хранения структуры, представляющей пакет протокола SNTP*/
			data->sntpdata = g_new0(sntpdg_t, 1);
			/*Установка стандартных значений заголовка пакета SNTP, отправляемого клиентом*/
			data->sntpdata->leapinfo = 0;
			data->sntpdata->version = 1;
			data->sntpdata->mode = 3;
			/*Получение текущей метки времени*/
			datetime = g_date_time_new_now_utc();
			/*Преобразование текущей метки времени в метку времени протокола SNTP*/
			if (g_date_time_to_timeval(datetime, &tv)) {
				/*В случае получения количества микросекунд*/
				data->sntpdata->transmit_timestamp_hi = GUINT32_TO_BE((guint32)(tv.tv_sec + 2208988800));
				data->sntpdata->transmit_timestamp_lo = GUINT32_TO_BE((guint32)(tv.tv_usec * 4295));
			} else {
				/*В том случае, если информация о микросекундах недоступна*/
				data->sntpdata->transmit_timestamp_hi = GUINT32_TO_BE((guint32)(g_date_time_to_unix(datetime) + 2208988800));
				data->sntpdata->transmit_timestamp_lo = 0;
			}
			/*Освобождение памяти, использованной для хранения текущей метки времени*/
			g_date_time_unref(datetime);
			/*Отправка сформированного пакета SNTP серверу*/
			len = g_socket_send_to(socket, socketaddress, (const gchar *)data->sntpdata, sizeof(sntpdg_t), NULL, &error);
			/*Уничтожение объекта адреса сетевого сокета*/
			g_object_unref(socketaddress);
			if (len != -1) {
				/*Создание нового источника событий на основе сетевого сокета с указанием условия доступности данных для получения*/
				source = g_socket_create_source(socket, G_IO_IN, NULL);
				/*Установка функции обратного вызова, которая будет вызываться при выполнении заданного условия*/
				g_source_set_callback(source, (GSourceFunc)application_read_callback, data, NULL);
				/*Соединение созданного источника событий с главным циклом обработки событий приложения*/
				g_source_attach(source, NULL);
				/*Удаление ссылки на созданный источник событий для его автоматического уничтожения*/
				g_source_unref(source);
			} else {
				application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
			}
		} else {
			application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
		}
		/*Уничтожение полученного списка IP-адресов*/
		g_resolver_free_addresses(addresses);
	} else {
		application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
	}
}

/*Функция запуска процесса синхронизации времени*/
static void application_start_sync_time(appdata *data)
{
	GResolver *resolver;
	const gchar *sntp_server;
		
	g_assert(data != NULL);
	
	/*Изменение состояния графического интерфейса*/
	application_process_ui_change(data, APPLICATION_PROCESS_START, NULL);
	
	/*Установка начальных значений*/
	data->timeoffset = 0;
	data->timesynced = FALSE;
	
	/*Получение полного доменного имени, соотвествующего IP-адресам группы серверов*/
	sntp_server = gtk_combo_box_get_active_id(GTK_COMBO_BOX(data->server_combo));
	
	if (sntp_server != NULL) {
		/*Создание объекта разрешения доменных имен и разрешение доменного имени*/
		resolver = g_resolver_get_default();
		g_resolver_lookup_by_name_async(resolver, sntp_server, NULL, application_resolver_lookup_callback, data);
	} else {
		application_process_ui_change(data, APPLICATION_PROCESS_ERROR, NULL);
	}
}

/*Функция, вызываемая при нажатии кнопки "Синхронизация"*/
static void application_sync_button_clicked(GtkButton *button, gpointer user_data)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	application_start_sync_time(data);
}

/*Функция таймера, предназначенная для обновления значений вермени*/
static gboolean application_upadate_time(gpointer user_data)
{
	appdata *data;
	GDateTime *systime, *synctime;
	gchar *timestr;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Получение текущей метки времени*/
	systime = g_date_time_new_now_local();
	
	if (systime != NULL) {
		/*Преобразование текущей метки времени в строку с разметкой*/
		timestr = g_date_time_format(systime, "<span size='35000'>%H:%M:%S</span>");
		if (timestr != NULL) {
			/*Вывод текущего времени и освобождение памяти, использованной для хранения строки*/
			gtk_label_set_markup(GTK_LABEL(data->systime_value), timestr);
			g_free(timestr);
		}
		/*Проверка установки флага синхронизации времени*/
		if (data->timesynced) {
			/*Получение синхронизированной метки времени путем коррекции текущей*/
			synctime = g_date_time_add(systime, data->timeoffset);
			if (synctime != NULL) {
				/*Преобразование синхронизированной метки времени в строку с разметкой*/
				timestr = g_date_time_format(synctime, "<span size='35000'>%H:%M:%S</span>");
				if (timestr != NULL) {
					/*Вывод синхронизированного времени и освобождение памяти, использованной для хранения строки*/
					gtk_label_set_markup(GTK_LABEL(data->synctime_value), timestr);
					g_free(timestr);
				}
				/*Освобождение памяти, использованной для хранения синхронизированной метки времени*/
				g_date_time_unref(synctime);
			}
		} else {
			/*Вывод шаблона времени в том случае, если флаг синхронизации не установлен*/
			gtk_label_set_markup(GTK_LABEL(data->synctime_value), "<span size='35000'>--:--:--</span>");
		}
		/*Освобождение памяти, использованной для хранения текущей метки времени*/
		g_date_time_unref(systime);
	}
	/*Значение TRUE позволяет повторно вызывать данную функцию*/
	return TRUE;
}

/*Функция для создания главного окна приложения*/
static void application_activate(GtkApplication *application, gpointer user_data)
{
	appdata *data;
	GtkWidget *vbox, *hbox;
	GtkWidget *systime_label, *synctime_label;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Создание главного окна*/
	data->mainwin = gtk_application_window_new(application);
	gtk_window_set_title(GTK_WINDOW(data->mainwin), "Пример использования класса GSocket (протокол UDP)");
	gtk_widget_set_size_request(data->mainwin, 600, 400);
	/*Создание контейнера для хранения элементов главного окна*/
	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
	/*Создание контейнера для элементов управления*/
	hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
	/*Создание виджета раскрывающегося списка для выбора группы серверов*/
	data->server_combo = gtk_combo_box_text_new();
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(data->server_combo), "europe.pool.ntp.org",        "Сервер из Европы");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(data->server_combo), "asia.pool.ntp.org",          "Сервер из Азии");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(data->server_combo), "north-america.pool.ntp.org", "Сервер из Северной америки");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(data->server_combo), "south-america.pool.ntp.org", "Сервер из Южной америки");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(data->server_combo), "oceania.pool.ntp.org",       "Сервер из Океании");
	gtk_combo_box_text_append(GTK_COMBO_BOX_TEXT(data->server_combo), "africa.pool.ntp.org",        "Сервер из Африки");
	/*Выбор первого элемента списка (с индексом 0)*/
	gtk_combo_box_set_active(GTK_COMBO_BOX(data->server_combo), 0);
	gtk_box_pack_start(GTK_BOX(hbox), data->server_combo, TRUE, TRUE, 5);
	/*Создание виджета кнопки для синхронизации времени*/
	data->sync_button = gtk_button_new_with_label("Синхронизация");
	g_signal_connect(data->sync_button, "clicked", G_CALLBACK(application_sync_button_clicked), data);
	gtk_box_pack_start(GTK_BOX(hbox), data->sync_button, FALSE, TRUE, 5);
	/*Добавление контейнера с элементами управления в основной контейнер*/
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 5);
	/*Создание виджета строки для вывода текущего времени и соотвествующего виджета строки для вывода заголовка*/
	systime_label = gtk_label_new("Системное время");
	gtk_box_pack_start(GTK_BOX(vbox), systime_label, TRUE, TRUE, 5);
	data->systime_value = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(data->systime_value), "<span size='35000'>--:--:--</span>");
	gtk_box_pack_start(GTK_BOX(vbox), data->systime_value, TRUE, TRUE, 5);
	/*Создание виджета строки для вывода синхронизированного времени и соотвествующего виджета строки для вывода заголовка*/
	synctime_label = gtk_label_new("Синхронизированное время");
	gtk_box_pack_start(GTK_BOX(vbox), synctime_label, TRUE, TRUE, 5);
	data->synctime_value = gtk_label_new(NULL);
	gtk_label_set_markup(GTK_LABEL(data->synctime_value), "<span size='35000'>--:--:--</span>");
	gtk_box_pack_start(GTK_BOX(vbox), data->synctime_value, TRUE, TRUE, 5);
	/*Создание виджета строки состояния*/
	data->statusbar = gtk_statusbar_new();
	data->statusmsgid = 0;
	data->statusconid = gtk_statusbar_get_context_id(GTK_STATUSBAR(data->statusbar), "GSNTP");
	/*Создание виджета индикатора выполнения задачи*/
	data->spinner = gtk_spinner_new();
	/*Добавление виджета индикатора выполнения задачи в строку состояния*/
	gtk_box_pack_end(GTK_BOX(data->statusbar), data->spinner, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), data->statusbar, FALSE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER(data->mainwin), GTK_WIDGET(vbox));
	/*Показ главного окна приложения*/
	gtk_application_add_window(application, GTK_WINDOW(data->mainwin));
	gtk_widget_show_all(data->mainwin);
	/*Скрытие виджета индикатора выполнения задачи*/
	gtk_widget_hide(data->spinner);
	/*Создание таймера для вызова функции вывода времени через каждые 500 мс*/
	application_upadate_time(data);
	data->timeoutsrc = g_timeout_add(500, application_upadate_time, data);
}

/*Функция, вызываемая перед завершением работы приложения*/
static void application_shutdown(GtkApplication *application, gpointer user_data)
{
	appdata *data;
	
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Заврешение работы таймера, вызывающего функцию вывода времени*/
	g_source_remove(data->timeoutsrc);
}

int main(int argc, char **argv)
{
	appdata *data;
	gint res;
	
	/*Создание структуры для хранения данных приложения*/
	data = g_new0(appdata, 1);
	/*Создание объекта приложения*/
	data->app = gtk_application_new("org.gtk.GSNTP", G_APPLICATION_FLAGS_NONE);
	/*Установка обработчиков основных сигналов объекта приложения*/
	g_signal_connect(data->app, "activate", G_CALLBACK(application_activate), data);
	g_signal_connect(data->app, "shutdown", G_CALLBACK(application_shutdown), data);
	/*Запуск основного цикла обработки событий приложения*/
	res = g_application_run(G_APPLICATION(data->app), 0, NULL);
	/*Освобождение памяти перед завершением работы приложения*/
	g_free(data);
	
	return res;
}
